@@ -7,6 +7,8 @@
 #include <sys/stat.h>
 #include <sys/socket.h>
 #include <sys/select.h>
+#include <netinet/in.h>
+#include <netinet/tcp.h>
 #include <unistd.h>
 #include <errno.h>
 #include <fcntl.h>
@@ -131,6 +133,7 @@ int load_config (char *configfile, struct config *cfg,
   cfg->num_s3fetchers = 2;
   cfg->s3timeout = 10000; // ms
   cfg->s3ssl = 1;
+  cfg->s3_max_reqs_per_conn = 100;
 
   for (i = 0; i < sizeof(cfg->s3conns)/sizeof(cfg->s3conns[0]); i++) {
     cfg->s3conns[i].sock = -1;
@@ -177,8 +180,10 @@ int load_config (char *configfile, struct config *cfg,
         sscanf(line, " port %7[0-9]", cfg->port) ||
         sscanf(line, " workers %hu", &cfg->num_io_threads) ||
         sscanf(line, " fetchers %hu", &cfg->num_s3fetchers) ||
+        sscanf(line, " s3maxreqsperconn %hu", &cfg->s3_max_reqs_per_conn) ||
         sscanf(line, " s3timeout %u", &cfg->s3timeout) ||
         sscanf(line, " s3ssl %hhu", &cfg->s3ssl) ||
+        sscanf(line, " s3name %127s", cfg->s3name) ||
         sscanf(line, " s3bucket %127s", cfg->s3bucket) ||
         sscanf(line, " s3accesskey %127s", cfg->s3accesskey) ||
         sscanf(line, " s3secretkey %127s", cfg->s3secretkey))
@@ -278,10 +283,29 @@ int save_pidfile (char *pidfile)
   return 0;
 }
 
+int set_socket_options (int sock)
+{
+  int opt;
+
+  opt = 1;
+  if (setsockopt(sock, SOL_SOCKET, SO_KEEPALIVE, &opt, sizeof(opt)) == -1)
+    return -1;
+
+  opt = TCP_RMEM;
+  if (setsockopt(sock, SOL_SOCKET, SO_RCVBUF, &opt, sizeof(opt)) == -1)
+    return -1;
+
+  opt = TCP_WMEM;
+  if (setsockopt(sock, SOL_SOCKET, SO_SNDBUF, &opt, sizeof(opt)) == -1)
+    return -1;
+
+  return 0;
+}
+
 static int s3_connect (struct s3connection *conn, char const **errstr)
 {
   struct addrinfo hints, *result, *walk;
-  int res, yes;
+  int res;
 
   memset(&hints, 0, sizeof(hints));
   hints.ai_family = AF_UNSPEC;
@@ -295,15 +319,10 @@ static int s3_connect (struct s3connection *conn, char const **errstr)
 
   for (walk = result; walk != NULL; walk = walk->ai_next) {
     conn->sock = socket(walk->ai_family, walk->ai_socktype, 0);
-    if (conn->sock >= 0) {
-      yes = 1;
-      res = setsockopt(conn->sock, SOL_SOCKET, SO_KEEPALIVE, &yes,
-                       sizeof(yes));
-
-      if ((res == 0) &&
-          (connect(conn->sock, walk->ai_addr, walk->ai_addrlen) == 0))
-        break;
-    }
+    if ((conn->sock >= 0) &&
+        (set_socket_options(conn->sock) == 0) &&
+        (connect(conn->sock, walk->ai_addr, walk->ai_addrlen) == 0))
+      break;
 
     *errstr = strerror(errno);
     close(conn->sock);
@@ -409,6 +428,7 @@ struct s3connection *s3_get_conn (struct config *cfg, unsigned int *conn_num,
     }
 
     ret->host = cfg->s3hosts[host];
+    ret->name = (cfg->s3name[0] == '\0' ? cfg->s3hosts[host] : cfg->s3name);
     ret->port = cfg->s3ports[port];
     ret->bucket = cfg->s3bucket;
     ret->timeout = cfg->s3timeout;
@@ -419,8 +439,12 @@ struct s3connection *s3_get_conn (struct config *cfg, unsigned int *conn_num,
 
       if ((cfg->s3ssl != 0) && (s3_tls_setup(ret, errstr) != 0))
         goto NEXT2;
+
+      ret->remaining_reqs = cfg->s3_max_reqs_per_conn;
     }
 
+    ret->remaining_reqs--;
+
     return ret;
 
     NEXT2:
@@ -433,7 +457,7 @@ struct s3connection *s3_get_conn (struct config *cfg, unsigned int *conn_num,
 
 void s3_release_conn (struct s3connection *conn)
 {
-  if (conn->is_error != 0) {
+  if ((conn->is_error != 0) || (conn->remaining_reqs == 0)) {
     if (conn->is_ssl != 0) {
       gnutls_bye(conn->tls_sess, GNUTLS_SHUT_RDWR);
       gnutls_deinit(conn->tls_sess);
@@ -608,7 +632,7 @@ static int s3_start_req (struct config *cfg, struct s3connection *conn,
            "Date: %s\r\n"
            "User-Agent: s3blkdev\r\n"
            "Authorization: AWS %s:",
-           httpverb_to_string(verb), string_to_sign + url_start, conn->host,
+           httpverb_to_string(verb), string_to_sign + url_start, conn->name,
            date, cfg->s3accesskey);
 
   res = sha1_b64(cfg->s3secretkey, string_to_sign, header + strlen(header),
